-module(database).

-behaviour(gen_server).

-export([start_link/1, init/1, handle_cast/2, handle_call/3, terminate/2]).
-export([get_tick/0]).

-record(state, {database_file = none, database = none, tick_start = none}).

first_error([]) -> ok;
first_error([ok | Rest]) -> first_error(Rest);
first_error([{rowid, _} | Rest]) -> first_error(Rest);
first_error([Error | _]) -> Error.

sql_begin(Database, Script) ->
  ok = sqlite3:sql_exec(Database, "BEGIN;"),
  Result = sqlite3:sql_exec_script(Database, Script),
  case first_error(Result) of
    ok -> sqlite3:sql_exec(Database, "COMMIT;");
    _ ->
      ok = sqlite3:sql_exec(Database, "ROLLBACK;"),
      Result
  end.

start_link(DatabaseFile) ->
  gen_server:start_link({local, ?MODULE}, ?MODULE, [DatabaseFile], []).

open_db(DatabaseFile) ->
  {ok, DB} = sqlite3:open(event_db, [{file, DatabaseFile}]),
  ok = sqlite3:sql_exec(DB, "PRAGMA foreign_keys = ON;"),
  HT = case sqlite3:sql_exec(DB, "SELECT tick FROM match_states;") of
         [{columns, C1}, {rows, R1}] ->
           if length(R1) == 0 -> 0;
              true ->
                T1 = get_column("tick", C1, R1),
                lists:max(T1)
           end;
         _ -> 0
       end,
  HFT = case sqlite3:sql_exec(DB, "SELECT tick FROM finals_states;") of
          [{columns, C2}, {rows, R2}] ->
            if length(R2) == 0 -> 0;
               true ->
                 T2 = get_column("tick", C2, R2),
                 lists:max(T2)
            end;
          _ -> 0
        end,
  HighestTick = lists:max([HT, HFT]),
  TickStart = erlang:monotonic_time(millisecond) - HighestTick,
  {TickStart, DB}.

init([DatabaseFile]) ->
  Exists = filelib:is_regular(DatabaseFile),
  if Exists =:= true ->
       {TickStart, DB} = open_db(DatabaseFile),
       {ok, #state{database_file = DatabaseFile, database = DB, tick_start = TickStart}};
     true ->
       io:format("Failed to load database ~s~n", [DatabaseFile]),
       {ok, #state{database_file = none, database = none}}
  end.

init_db(Database) ->
  ok = sqlite3:create_table(Database, divisions, [{division, text, [not_null]}],
                            [{primary_key, [division]}]),

  ok = sqlite3:create_table(Database, fields, [{field, text, [not_null]},
                                               {serial, text, [unique]},
                                               {division, text},
                                               {other, text}],
                            [{primary_key, [field]},
                             {foreign_key, {[division], divisions, [division], "ON DELETE SET NULL"}},
                             {check, "other IN ('testing', 'skills')"}]),

  ok = sqlite3:create_table(Database, teams, [{team, text, [not_null]},
                                              {division, text},
                                              {inspection, blob}],
                            [{primary_key, [team]},
                             {foreign_key, {[division], divisions, [division], "ON DELETE SET NULL"}}]),

  ok = sqlite3:create_table(Database, matches, [{division, text, [not_null]},
                                                {type, text, [not_null]},
                                                {number, integer, [not_null]}],
                            [{primary_key, [division, type, number]},
                             {check, "type IN ('practice', 'qualification')"},
                             {foreign_key, {[division], divisions, [division], "ON DELETE CASCADE"}}]),

  ok = sqlite3:create_table(Database, match_states, [{division, text, [not_null]},
                                                     {type, text, [not_null]},
                                                     {number, integer, [not_null]},
                                                     {tick, integer, [not_null]},
                                                     {state, text, [not_null]}],
                            [{primary_key, [division, type, number, tick]},
                             {foreign_key, {[division, type, number], matches, [division, type, number], "ON DELETE CASCADE"}}]),

  ok = sqlite3:create_table(Database, match_teams, [{division, text, [not_null]},
                                                    {type, text, [not_null]},
                                                    {number, integer, [not_null]},
                                                    {side, text, [not_null]}, 
                                                    {anum, integer, [not_null]},
                                                    {team, text, [not_null]}],
                            [{primary_key, [division, type, number, side, anum]},
                             {check, "side IN ('red', 'blue')"},
                             {check, "anum >= 1 AND anum <= 2"},
                             {check, "type IN ('practice', 'qualification')"},
                             {foreign_key, {[division, type, number], matches, [division, type, number], "ON DELETE CASCADE"}},
                             {foreign_key, {[team], teams, [team], ""}}]),

  ok = sqlite3:create_table(Database, match_scores, [{division, text, [not_null]},
                                                     {type, text, [not_null]},
                                                     {number, integer, [not_null]},
                                                     {history, integer, [not_null]},
                                                     {uuid, integer, [not_null]}],
                            [{primary_key, [division, type, number, history]},
                             {foreign_key, {[uuid], scores, [uuid], "ON DELETE CASCADE"}},
                             {foreign_key, {[division, type, number],
                                            matches,
                                            [division, type, number],
                                            "ON DELETE CASCADE"}}]),

  ok = sqlite3:create_table(Database, brackets, [{division, text},
                                                 {number, integer, [not_null]}],
                            [{primary_key, [division, number]},
                             {foreign_key, {[division], divisions, [division], "ON DELETE CASCADE"}}]),

  ok = sqlite3:create_table(Database, bracket_states, [{division, text},
                                                       {number, integer, [not_null]},
                                                       {tick, integer, [not_null]},
                                                       {state, text, [not_null]}],
                            [{primary_key, [division, number, tick]},
                             {foreign_key, {[division, number], brackets, [division, number], "ON DELETE CASCADE"}}]),

  ok = sqlite3:create_table(Database, elimination_alliances, [{division, text, [not_null]},
                                                              {number, integer, [not_null]},
                                                              {side, text, [not_null]},
                                                              {anum, integer, [not_null]}],
                            [{primary_key, [division, number, side]},
                             {check, "side IN ('red', 'blue')"},
                             {foreign_key, {[division, number], brackets, [division, number], "ON DELETE CASCADE"}},
                             {foreign_key, {[division, anum], alliances, [division, number], ""}}]),

  ok = sqlite3:create_table(Database, finals_alliances, [{division, text},
                                                         {number, integer, [not_null]},
                                                         {side, text, [not_null]},
                                                         {adiv, text, [not_null]}],
                            [{primary_key, [division, number, side]},
                             {check, "side IN ('red', 'blue')"},
                             {check, "division IS NULL"},
                             {foreign_key, {[division, number], brackets, [division, number], "ON DELETE CASCADE"}},
                             {foreign_key, {[adiv], winners, [division], ""}}]),

  ok = sqlite3:create_table(Database, bracket_scores, [{division, text},
                                                       {number, integer, [not_null]},
                                                       {instance, integer, [not_null]},
                                                       {history, integer, [not_null]},
                                                       {uuid, integer, [not_null]}],
                            [{primary_key, [division, number, instance, history]},
                             {foreign_key, {[uuid], scores, [uuid], "ON DELETE CASCADE"}},
                             {foreign_key, {[division, number],
                                            brackets,
                                            [division, number],
                                            "ON DELETE CASCADE"}}]),

  ok = sqlite3:create_table(Database, scores, [{uuid, integer, [not_null]},
                                               {autonomous, integer, [not_null]},
                                               {red_elevation_1, integer, [not_null]},
                                               {red_elevation_2, integer, [not_null]},
                                               {red_goal, integer, [not_null]},
                                               {red_zone, integer, [not_null]},
                                               {red_agoal, integer, [not_null]},
                                               {red_azone, integer, [not_null]},
                                               {red_1, text},
                                               {red_2, text},
                                               {blue_elevation_1, integer, [not_null]},
                                               {blue_elevation_2, integer, [not_null]},
                                               {blue_goal, integer, [not_null]},
                                               {blue_zone, integer, [not_null]},
                                               {blue_agoal, integer, [not_null]},
                                               {blue_azone, integer, [not_null]},
                                               {blue_1, text},
                                               {blue_2, text}],
                            [{primary_key, [uuid]},
                             {check, "red_1 IS NULL OR red_1 IN ('dq', 'no show')"},
                             {check, "red_2 IS NULL OR red_2 IN ('dq', 'no show')"},
                             {check, "blue_1 IS NULL OR blue_1 IN ('dq', 'no show')"},
                             {check, "blue_2 IS NULL OR blue_2 IN ('dq', 'no show')"}]),

  ok = sqlite3:create_table(Database, alliances, [{division, text, [not_null]},
                                                  {number, integer, [not_null]}],
                            [{primary_key, [division, number]},
                             {foreign_key, {[division], divisions, [division], "ON DELETE CASCADE"}}]),

  ok = sqlite3:create_table(Database, alliance_members, [{division, text, [not_null]},
                                                         {number, integer, [not_null]},
                                                         {anum, integer, [not_null]},
                                                         {team, text, [not_null]}],
                            [{primary_key, [division, number, anum]},
                             {foreign_key, {[division, number], alliances, [division, number], "ON DELETE CASCADE"}},
                             {foreign_key, {[team], teams, [team], "ON DELETE CASCADE"}}]),

  ok = sqlite3:create_table(Database, winners, [{division, text, [not_null]},
                                                {alliance, integer, [not_null]}],
                            [{primary_key, [division]},
                             {foreign_key, {[division, alliance], alliances, [division, number], "ON DELETE CASCADE"}}]),

  ok = sqlite3:create_table(Database, skills_scores, [{team, text, [not_null]},
                                                      {type, text, [not_null]},
                                                      {attempt, integer, [not_null]},
                                                      {elevation, integer, [not_null]},
                                                      {goal, integer, [not_null]},
                                                      {zone, integer, [not_null]},
                                                      {agoal, integer, [not_null]},
                                                      {azone, integer, [not_null]}],
                            [{primary_key, [team, type, attempt]},
                             {check, "type IN ('driver', 'autonomous')"},
                             {foreign_key, {[team], teams, [team], ""}}]),
  ok.

delete_teams(Database, Teams) ->
  SQL = lists:append([io_lib:format("DELETE FROM teams WHERE team = ~p; ",
                                    [Team])
                      || Team <- Teams]),
  sql_begin(Database, SQL).

add_teams(Database, Teams) ->
  SQL = lists:append([io_lib:format("INSERT INTO teams(team) VALUES('~s') ON CONFLICT(team) DO NOTHING; ",
                                    [Team])
                      || Team <- Teams]),
  sql_begin(Database, SQL).

assign_divisions(Database, Teams) ->
  SQL = lists:append([io_lib:format("UPDATE teams SET division = '~s' WHERE team = '~s'; ",
                                    [Division, Team])
                      || {Team, Division} <- Teams]),
  sql_begin(Database, SQL).

write_division_matches(Database, Division, Type, Matches) ->
  Round = atom_to_list(Type),
  SQL = lists:append([io_lib:format("DELETE FROM matches WHERE division = '~s' AND type = '~s'; ",
                                    [Division, Round]),
                      lists:append([io_lib:format(
                                      "INSERT INTO matches(division, type, number) VALUES ('~s', '~s', ~p);
                                      INSERT INTO match_teams(division, type, number, side, anum, team) VALUES ('~s', '~s', ~p, 'red',  1, '~s');
                                       INSERT INTO match_teams(division, type, number, side, anum, team) VALUES ('~s', '~s', ~p, 'red',  2, '~s');
                                       INSERT INTO match_teams(division, type, number, side, anum, team) VALUES ('~s', '~s', ~p, 'blue', 1, '~s');
                                       INSERT INTO match_teams(division, type, number, side, anum, team) VALUES ('~s', '~s', ~p, 'blue', 2, '~s');",
                                      [Division, Round, N, Division, Round, N, B1, Division, Round, N, B2, Division, Round, N, R1, Division, Round, N, R2])
                                    || {N, [B1, B2, R1, R2]} <- lists:enumerate(Matches)])]),
  sql_begin(Database, SQL).

index_of(Item, List) -> index_of(Item, List, 1).
index_of(Item, [Element | _], N) when Item == Element -> N;
index_of(Item, [_ | List], N) -> index_of(Item, List, N+1).

get_column(_, [], _) -> [];
get_column(_, _, []) -> [];
get_column(Name, Columns, Rows) ->
  Index = index_of(Name, Columns),
  [element(Index, Row) || Row <- Rows].

handle_call({time_sync, T0}, _, State) ->
  {reply, {T0, get_tick(State)}, State};
handle_call({new_db, DatabaseFile}, _, State) ->
  ok = if State#state.database =:= none -> ok;
          true -> sqlite3:close(State#state.database)
       end,
  Exists = filelib:is_regular(DatabaseFile),
  ok = if Exists -> file:delete(DatabaseFile);
          true -> ok
       end,
  {TickStart, Database} = open_db(DatabaseFile),
  {reply, init_db(Database), State#state{database_file = DatabaseFile, database = Database, tick_start = TickStart}};
handle_call({load_db, DatabaseFile}, _, State) ->
  true = filelib:is_regular(DatabaseFile),
  {TickStart, Database} = if DatabaseFile == State#state.database_file ->
                               {State#state.tick_start, State#state.database};
                             true ->
                               ok = if State#state.database =:= none -> ok;
                                       true -> sqlite3:close(State#state.database)
                                    end,
                               open_db(DatabaseFile)
                          end,

  % Publish division/{division} with fields
  [_, {rows, Divisions}] = sqlite3:sql_exec(Database, "SELECT division FROM divisions;"),
  [_, {rows, DivisionFields}] = sqlite3:sql_exec(Database, "SELECT field, division FROM fields WHERE division IS NOT NULL;"),
  DivisionMessages = [{list_to_binary(io_lib:format("division/~s", [Name])),
                       jsone:encode(#{<<"fields">> =>
                                      lists:filtermap(fun({Field, FieldDivision}) ->
                                                          if FieldDivision == Name ->
                                                               {'true', Field};
                                                             true -> false
                                                          end
                                                      end, DivisionFields)
                                     })}
                      || {Name} <- Divisions],

  % Publish team/{team} with inspection
  % Publish division/{division}/teams with team list
  [_, {rows, Teams}] = sqlite3:sql_exec(Database, "SELECT team, division, inspection FROM teams"),
  TeamsMessages = [{list_to_binary(io_lib:format("team/~s", [Team])),
                    jsone:encode(#{<<"inspection">> => Inspection})}
                   || {Team, _, Inspection} <- Teams],
  DivisionTeams = [{list_to_binary(io_lib:format("division/~s/teams", [Name])),
                    jsone:encode(lists:filtermap(fun({Team, Div, _}) ->
                                                     if Div == Name ->
                                                          {'true', Team};
                                                        true ->
                                                          false
                                                     end
                                                 end, Teams))}
                   || {Name} <- Divisions],

  % Read matches and match_teams tables and publish to match/{div}/{round}/{number}
  % Read match_sates and publish to match/{div}/{round}/{number}/state
  % Read match_scores and publish to match/{div}/{round}/{number}/state
  % Read rankings and publish divisions/{div}/rankings
  % Read skills scores and publish team/{team}/skills
  % Read alliances and publish to division/{div}/alliances


  % DEBUG PRINTS
  io:fwrite("DIVISIONS: ~p~n", [DivisionMessages]),
  io:fwrite("TEAMS: ~p~n", [TeamsMessages]),
  io:fwrite("INSPECTION: ~p~n", [DivisionTeams]),

  {reply, ok, State#state{database_file = DatabaseFile, database = Database, tick_start = TickStart}};
handle_call(close_db, _, State) ->
  if State#state.database =:= none -> {reply, ok, State};
     true ->
       ok = sqlite3:close(State#state.database),
       {reply, ok, State#state{database = none, database_file = none}}
  end;

handle_call({add_teams, Teams}, _, State) ->
  ok = add_teams(State#state.database, Teams),
  {reply, ok, State};
handle_call({assign_divisions, Teams}, _, State) ->
  ok = assign_divisions(State#state.database, Teams),
  {reply, ok, State};
handle_call(split_teams, _, State) ->
  Teams = [X||{_,X} <- lists:sort([{rand:uniform(), N} || N <- get_teams(State#state.database)])],
  [{columns, Cols}, {rows, Rows}] = sqlite3:sql_exec(State#state.database, "SELECT division FROM divisions;"),
  Divisions = get_column("division", Cols, Rows),
  Assignments = [{Team, lists:nth((I rem length(Divisions)) + 1, Divisions)}
                 || {I, Team} <- lists:enumerate(Teams)],
  ok = assign_divisions(State#state.database, Assignments),
  {reply, ok, State};
handle_call({delete_teams, Removed}, _, State) ->
  ok = delete_teams(State#state.database, Removed),
  {reply, ok, State};

handle_call({match_score, Division, Round, Number, Instance, Red, Blue, Score}, _, State) ->
  [{columns, Columns}, {rows, Rows}] = sqlite3:sql_exec(State#state.database,
                                                        io_lib:format(
                                                          "SELECT history FROM match_scores WHERE Division = ~p AND type = '~s' AND number = ~p AND instance = ~p;",
                                                          [Division, atom_to_list(Round), Number, Instance])),
  Instances = get_column("history", Columns, Rows),
  NextInstance = if length(Instances) == 0 -> 1;
                    true -> lists:max(Instances) + 1
                 end,
  ok = first_error(sqlite3:sql_exec_script(State#state.database, lists:append(
                                                                   [io_lib:format("INSERT INTO match_scores(division, type, number, instance, history, red, blue, score) VALUES(~p, '~s', ~p, ~p, ~p, ~p, ~p, '~s');", [Division, atom_to_list(Round), Number, Instance, NextInstance, Red, Blue, Score]),
                                                                    io_lib:format("INSERT INTO match_states(division, type, number, tick, state) VALUES(~p, '~s', ~p, ~p, '~s');", [Division, atom_to_list(Round), Number, get_tick(State), "scored"])]))),
  {reply, ok, State};
handle_call({match_state, Division, Round, Number, MatchState}, _, State) ->
  {rowid, _} = sqlite3:sql_exec(State#state.database,
                                io_lib:format("INSERT INTO match_states(division, type, number, tick, state) VALUES(~p, '~s', ~p, ~p, '~s');",
                                              [Division, atom_to_list(Round), Number, get_tick(State), MatchState])),
  {reply, ok, State};

handle_call({add_division, Name}, _, State) ->
  {rowid, _} = sqlite3:sql_exec(
                 State#state.database,
                 io_lib:format("INSERT INTO divisions(division) VALUES('~s');",
                               [Name])),
  {reply, ok, State};
handle_call({delete_division, Division}, _, State) ->
  ok = first_error(sqlite3:sql_exec_script(
                     State#state.database,
                     io_lib:format(
                       "UPDATE teams SET division = NULL WHERE division = '~s';
                       DELETE FROM divisions WHERE division = '~s';",
                       [Division, Division]))),
  {reply, ok, State};
handle_call({generate_practice, Division, Size}, _, State) ->
  Teams = get_div_teams(State#state.database, Division),
  Resp = case schedule:create(rand:uniform(10000000), Size, Teams) of
           error -> error;
           Matches -> write_division_matches(State#state.database, Division, practice, Matches)
         end,
  {reply, Resp, State};
handle_call({generate_qualification, Division, Size}, _, State) ->
  Teams = get_div_teams(State#state.database, Division),
  Resp = case schedule:create(rand:uniform(10000000), Size, Teams) of
           error -> error;
           Matches -> write_division_matches(State#state.database, Division, qualification, Matches)
         end,
  {reply, Resp, State};
handle_call({generate_elimination, Division, Size}, _, State) ->
  Bracket = schedule:generate_bracket(Size),
  SQL = lists:append(
          [
           io_lib:format("DELETE FROM alliances WHERE division = '~s'; ", [Division]),
           io_lib:format("DELETE FROM brackets WHERE division = '~s'; ", [Division]),
           lists:append([io_lib:format("INSERT INTO alliances(division, number) VALUES('~s', ~p); ",
                                       [Division, N])
                         || N <- lists:seq(1, Size)]),
           lists:append([io_lib:format("INSERT INTO brackets(division, number)
                                       VALUES('~s', ~p);",
                                           [Division, N])
                                           || N <- lists:seq(1, Size-1)])]),

  {reply, sql_begin(State#state.database, SQL), State};
handle_call(generate_finals, _, State) ->
  {reply, noimpl, State};
handle_call(_, _, State) ->
  {reply, noimpl, State}.

get_div_teams(Database, Division) ->
  [{columns, Columns}, {rows, Rows}] = sqlite3:sql_exec(Database, io_lib:format("SELECT team FROM teams WHERE division = ~p;", [Division])),
  get_column("team", Columns, Rows).

get_teams(Database) ->
  [{columns, Columns}, {rows, Rows}] = sqlite3:sql_exec(Database, "SELECT team FROM teams;"),
  get_column("team", Columns, Rows).

handle_cast(_, State) ->
  {noreply, State}.

get_tick(State) -> erlang:monotonic_time(millisecond) - State#state.tick_start.
get_tick() ->
  T0 = erlang:monotonic_time(millisecond),
  {T0, T1} = gen_server:call(?MODULE, {time_sync, T0}),
  T2 = erlang:monotonic_time(millisecond),
  Offset = trunc(((T1-T0) + (T1-T2))/2),
  NetworkDelay = trunc((T2-T0)/2),
  ServerTick = T2 + Offset + NetworkDelay,
  {Offset, ServerTick}.

terminate(_, State) ->
  DB = State#state.database,
  ok = if DB =:= none -> ok; true -> sqlite3:close(DB) end.
