-module(schedule).

-behaviour(gen_server).

-export([start_link/0, init/1, handle_cast/2, handle_call/3]).

-export([create/3, generate_bracket/1, parent/3]).

-include("schedule.hrl").

higher_pow2(N) -> round(math:pow(2, ceil(math:log2(N)))).

first_layer(BracketSize, _, N) when N > BracketSize/2 -> [];
first_layer(BracketSize, Size, N) when BracketSize - N + 1 > Size ->
  [N | first_layer(BracketSize, Size, N+1)];
first_layer(BracketSize, Size, N) ->
  [{N, BracketSize - N + 1} | first_layer(BracketSize, Size, N+1)].

first_layer(Size) when Size > 1 ->
  BracketSize = higher_pow2(Size),
  first_layer(BracketSize, Size, 1).

merge_layer([Element], []) -> Element;
merge_layer([], Merged) ->
  Final = lists:reverse(Merged),
  merge_layer(Final);
merge_layer([First | Rest], Merged) ->
  Last = lists:last(Rest),
  List = lists:droplast(Rest),
  merge_layer(List, [{First, Last} | Merged]).
merge_layer(List) -> merge_layer(List, []).

label_bracket({Left, Right}, [N | Counters], Depth) ->
  {LC, L} = label_bracket(Left, Counters, Depth+1),
  {RC, R} = label_bracket(Right, LC, Depth+1),
  {[N+1 | RC], #bracket{num=N, red=L, blue=R}};
label_bracket(Elem, Counters, _) -> {Counters, Elem}.

label_bracket(Bracket, Depth) ->
  {_, Labeled} = label_bracket(Bracket, lists:duplicate(Depth, 1), 1),
  Labeled.

generate_bracket(Size) ->
  label_bracket(merge_layer(first_layer(Size)), ceil(math:log2(Size))).

parent(Depth, Level, Index, #bracket{num = N, red = #bracket{num=Index}}) when Depth == Level - 1 ->
  {N, red};
parent(Depth, Level, Index, #bracket{num = N, blue = #bracket{num=Index}}) when Depth == Level - 1 ->
  {N, blue};
parent(Depth, Level, _, #bracket{}) when Depth >= Level -> false;
parent(Depth, Level, Index, #bracket{red = Left, blue = Right}) ->
  L = parent(Depth + 1, Level, Index, Left),
  if L =:= false -> parent(Depth + 1, Level, Index, Right);
     true -> L
  end.

parent(Level, Index, Root) -> parent(1, Level, Index, Root).

create(Seed, Size, Teams) ->
  gen_server:call(?MODULE, {generate, Seed, Size, Teams}).

start_link() ->
  gen_server:start_link({local, ?MODULE}, ?MODULE, [], []).

init([]) ->
  {ok, []}.

handle_cast(_, State) ->
  {noreply, State}.

handle_call({generate, Seed, Size, Teams}, _, State) ->
  {reply, create_schedule(Seed, Size, Teams), State}.

duplicate(1, Original, Constructed) -> lists:append(Original, Constructed);
duplicate(N, Original, Constructed) -> duplicate(N-1, Original, lists:append(Original, Constructed)).

duplicate(List, N) -> duplicate(List, N, []).

padding(Size) when Size >= 4 -> (4 - (Size rem 4)) rem 4.

fill_padding_match([], Teams, Filled) -> {Filled, Teams};
fill_padding_match([padding | Left], [Picked | Teams], Filled) ->
  Duplicate = lists:member(Picked, Left) or lists:member(Picked, Filled),
  if Duplicate -> fill_padding_match([padding | Left], lists:append(Teams, [Picked]), Filled);
     true -> fill_padding_match(Left, Teams, [Picked | Filled])
  end;
fill_padding_match([Team | Left], Teams, Filled) ->
  fill_padding_match(Left, Teams, [Team | Filled]).

fill_padding_match(Match, Teams) ->
  fill_padding_match(Match, Teams, []).

fill_padding_schedule(Matches, Teams) ->
  TeamsRandom = [X || {_, X} <- lists:sort([{rand:uniform(), N} || N <- Teams])],
  {Schedule, _} = lists:mapfoldl(fun fill_padding_match/2, TeamsRandom, Matches),
  Schedule.

create_schedule(Seed, Size, Teams) ->
  TeamsRepeated = duplicate(Size, Teams),
  TeamsPadded = TeamsRepeated ++ lists:duplicate(padding(length(Teams)*Size), padding),
  rand:seed(default, Seed),
  TeamsRandom = [X || {_, X} <- lists:sort([{rand:uniform(), N} || N <- TeamsPadded])],
  case pick_matches(TeamsRandom) of
    error -> error;
    Matches -> fill_padding_schedule(Matches, Teams)
  end.

pick_teams(_, _, _, 0) -> error;
pick_teams(Teams, Picked, 0, Safety) when Safety > 0 -> {Teams, Picked};
pick_teams([Team | Teams], Picked, N, Safety) when Safety > 0 ->
  Duplicate = lists:member(Team, Picked),
  if Duplicate, Team =/= padding -> pick_teams(lists:append(Teams, [Team]), Picked, N, Safety - 1);
     true -> pick_teams(Teams, [Team | Picked], N-1, Safety)
  end.

pick_teams(Teams, N) -> pick_teams(Teams, [], N, 4*length(Teams)).

pick_matches([], Matches) -> Matches;
pick_matches(Teams, Matches) ->
  case pick_teams(Teams, 4) of
    {NewTeams, Match} -> pick_matches(NewTeams, [Match | Matches]);
    error -> error
  end.

pick_matches(Teams) -> pick_matches(Teams, []).
